/*
 * Copyright 2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.ws.resource;

import javax.wsdl.Definition;
import javax.wsdl.PortType;
import javax.xml.namespace.QName;
import java.util.Map;

/**
 * A WSRF resource capability (i.e. set of operations and properties).
 *
 * @author Ian Springer
 */
public interface ResourceCapability
{

    /**
     * @return the names of all custom operations (i.e. operations that are not derived from a "base"
     *         ResourceDefinition)
     */
    String[] getCustomOperationNames();

    /**
     * @return the names of all custom properties (i.e. proeprties that are not derived from a "base"
     *         ResourceDefinition)
     */
    QName[] getCustomPropertyNames();

    /**
     * Returns the JWSDL {@link Definition} for this WS-Resource capability.
     *
     * @return
     */
    Definition getDefinition();

    /**
     * Returns the JWSDL {@link PortType} corresponding to this WS-Resource capability.
     *
     * @return
     */
    PortType getPortType();

    /**
     * @return a map of any other {@link ResourceCapability}s that this capability implements;
     *         the map's keys are the {@link QName}s of the capability portTypes, and the values
     *         are the ResourceCapabilitys.
     */
    Map getImplementedResourceCapabilities();

    /**
     *
     * @param capabilityName
     *
     * @return
     */
    boolean implementsResourceCapability( QName capabilityName );

    /**
     * Returns a list of the QNames of the resource properties that are defined for the specified portType, or null if
     * the portType does not have any associated resource properties.
     *
     * @return the QNames of the resource properties defined for the specified portType
     */
    QName[] getPropertyNames();

    /**
     * @return true if this resource defines a resource properties document, or false if it does not
     */
    boolean hasProperties();

    /**
     * Returns the name of the resource properties document element, as specified by
     * the wsrf-rp:ResourceProperties attribute on the portType, or null if this
     * resource does not define a properties document.
     *
     * @return
     */
    QName getPropertiesDocumentName();

    /**
     * Returns the name of the WS-RMD metadata descriptor, as specified by
     * the wsrmd:MetadataDescriptor attribute on the portType, or null if this
     * resource does not define a metadata descriptor.
     *
     * @return
     */
    QName getMetadataDescriptorName();

    /**
     * Returns the location of the WS-RMD metadata definitions document, as specified by
     * the wsrmd:MetadataDescriptorLocation attribute on the portType, or null if this
     * resource does not define a metadata descriptor.
     *
     * @return
     */
    String getMetadataDescriptorLocation();

}
