/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource;

import org.apache.ws.resource.faults.ResoureKeyHeaderNotFoundFaultException;

import javax.xml.namespace.QName;
import javax.xml.rpc.handler.soap.SOAPMessageContext;
import javax.xml.soap.SOAPHeaderElement;
import javax.xml.soap.SOAPMessage;
import java.net.URL;
import java.util.Iterator;

/**
 * LOG-DONE
 * <code>ResourceContext</code> wraps around a {@link SOAPMessageContext SOAPMessageContext} and provides convenience
 * functions for obtaining {@link ResourceKey ResourceKey}, {@link ResourceHome ResourceHome}, and resource objects
 * associated with the given SOAP message (referenced in <code>SOAPMessageContext</code>).
 */
public abstract class ResourceContext
{
   /**
    * Creates a new {@link ResourceContext} object.
    */
   protected ResourceContext(  )
   {
   }

   /**
    * @see SOAPMessageContext#setProperty(String, Object)
    */
   public abstract void setProperty( String name,
                                     Object value );

   /**
    * @see SOAPMessageContext#getProperty(String)
    */
   public abstract Object getProperty( String name );

   /**
    * @see SOAPMessageContext#getPropertyNames()
    */
   public abstract Iterator getPropertyNames(  );

   /**
    * Gets actual resource.
    *
    * @return The actual resource object. Cannot be null.
    *
    * @throws ResourceUnknownException  if resource was not specified in the request or it does not exist.
    * @throws ResourceContextException if error occurs during resource lookup.
    */
   public abstract Resource getResource(  )
   throws ResourceContextException, 
          ResourceException;

   /**
    * Gets <code>ResourceHome</code> associated with the service.
    *
    * @return <code>ResourceHome</code> associated with the service. Cannot be null.
    *
    * @throws NoResourceHomeException  if <code>ResourceHome</code> is not configured for the specified service.
    * @throws ResourceContextException if any other error.
    */
   public abstract ResourceHome getResourceHome(  )
   throws ResourceContextException;

   /**
    * Gets <code>ResourceKey</code> of a given name and type.
    *
    * @return <code>ResourceKey</code> of the given type and name. Returns null if the key of the specified name was not
    *         found in the message.
    *
    * @throws ResourceContextException if error occurs during lookup.
    */
   public abstract ResourceKey getResourceKey( QName keyName,
                                               Class keyClass )
   throws ResourceContextException, 
          ResoureKeyHeaderNotFoundFaultException;

   /**
    * Gets <code>SOAPHeaderElement</code> that contains the resource key for the service using <code>ResourceHome</code>
    * lookup.
    *
    * @return The <code>SOAPHeaderElement</code> that contains the resource key for the service. Returns null if one not
    *         found.
    *
    * @throws ResourceContextException if error occurs during lookup.
    * @see #getResourceHome()
    */
   public abstract SOAPHeaderElement getResourceKeyHeader(  )
   throws ResourceContextException;

   /**
    * Gets <code>SOAPHeaderElement</code> of the specified name that contains the resource key.
    *
    * @return The <code>SOAPHeaderElement</code> that contains the resource key of the given name. Returns null if one
    *         not found.
    *
    * @throws ResourceContextException if error occurs during lookup.
    */
   public abstract SOAPHeaderElement getResourceKeyHeader( QName keyName )
   throws ResourceContextException;

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public abstract SOAPMessage getSOAPMessage(  );

   /**
    * Returns <code>ResourceContext</code> initialized with a specific <code>SOAPMessageContext</code>.
    *
    * @return <code>ResourceContext</code> initialized with a specific <code>SOAPMessageContext</code>.
    */
   public static ResourceContext getResourceContext( SOAPMessageContext soapMsgContext )
   {
      // TODO: return ResourceContext impl based using configurable factory.
      return null;
   }

   /**
    * Gets <code>ResourceKey</code> instance using <code>ResourceHome</code> lookup.
    *
    * @return <code>ResourceKey</code> of name and type associated with the target service' <code>ResourceHome</code>.
    *         Returns null if key is not found in the message.
    *
    * @throws ResourceContextException if error occurs during lookup.
    * @see #getResourceHome()
    */
   public abstract ResourceKey getResourceKey(  )
   throws ResourceContextException;

   /**
    * Gets path of the service.
    *
    * @return The path of the service.
    */
   public abstract String getServiceName(  );

   /**
    * Gets the full URL of the service
    *
    * @return The URL of the service
    */
   public abstract URL getServiceURL(  );

   /**
    * Gets the base URL from which service urls are based.
    *
    * @return The String representation of the base url
    */
   public abstract String getBaseURL(  );

   /**
    * @see SOAPMessageContext#containsProperty(String)
    */
   public abstract boolean containsProperty( String name );

   /**
    * @see SOAPMessageContext#removeProperty(String)
    */
   public abstract void removeProperty( String name );
}