/*=============================================================================*
 *  Copyright 2005 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.faults;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.ws.Soap1_1Constants;
import org.apache.ws.resource.i18n.Keys;
import org.apache.ws.resource.i18n.MessagesImpl;
import org.apache.ws.resource.properties.NamespaceVersionHolder;
import org.apache.ws.util.XmlBeanNameUtils;
import org.apache.ws.util.XmlBeanUtils;
import org.apache.ws.util.i18n.Messages;
import org.apache.xmlbeans.XmlObject;

import javax.xml.namespace.QName;
import javax.xml.rpc.JAXRPCException;
import javax.xml.soap.Detail;
import javax.xml.soap.DetailEntry;
import java.lang.reflect.Method;
import java.util.Calendar;

/**
 * The WS-BaseFaults base fault type implemented as a Java exception.
 *
 * @author Ian P. Springer
 */
public abstract class AbstractBaseFaultException extends FaultException
{

    private static final Messages MSG = MessagesImpl.getInstance();

    private static final QName DEFAULT_FAULT_CODE = Soap1_1Constants.FAULT_CLIENT;

    private Calendar m_timestamp;
    private EndpointReferenceType m_originator; // TODO: use our EPR facade instead of the Apache Addressing type
    private ErrorCode m_errorCode;
    private Description[] m_description;
    private AbstractBaseFaultException[] m_faultCause;
    private NamespaceVersionHolder m_namespaceSet;
    private DetailEntry m_detailEntry;

    public AbstractBaseFaultException( NamespaceVersionHolder namespaceSet, String faultString, QName faultCode )
    {
        super( faultCode, faultString );
        m_namespaceSet = namespaceSet;
        m_timestamp = Calendar.getInstance();
    }

    public AbstractBaseFaultException( NamespaceVersionHolder namespaceSet, String faultString )
    {
        this( namespaceSet, faultString, DEFAULT_FAULT_CODE );
    }

    /**
     * DOCUMENT_ME
     *
     * @param description DOCUMENT_ME
     */
    public void setDescription( Description[] description )
    {
        m_description = description;
    }

    /**
     * DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    public Description[] getDescription()
    {
        return m_description;
    }

    /**
     * DOCUMENT_ME
     *
     * @param errorCode DOCUMENT_ME
     */
    public void setErrorCode( ErrorCode errorCode )
    {
        m_errorCode = errorCode;
    }

    /**
     * DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    public ErrorCode getErrorCode()
    {
        return m_errorCode;
    }

    /**
     * DOCUMENT_ME
     *
     * @param faultCause DOCUMENT_ME
     */
    public void setFaultCause( AbstractBaseFaultException[] faultCause )
    {
        m_faultCause = faultCause;
    }

    /**
     * DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    public AbstractBaseFaultException[] getFaultCause()
    {
        return m_faultCause;
    }

    /**
     * DOCUMENT_ME
     *
     * @param originator DOCUMENT_ME
     */
    public void setOriginator( EndpointReferenceType originator )
    {
        m_originator = originator;
    }

    /**
     * DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    public EndpointReferenceType getOriginator()
    {
        return m_originator;
    }

    /**
     * DOCUMENT_ME
     *
     * @param timestamp DOCUMENT_ME
     */
    public void setTimestamp( Calendar timestamp )
    {
        m_timestamp = timestamp;
    }

    /**
     * DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    public Calendar getTimestamp()
    {
        return m_timestamp;
    }

    public NamespaceVersionHolder getNamespaceSet()
    {
        return m_namespaceSet;
    }

    public abstract QName getBaseFaultName();

    public Detail getDetail()
    {
        updateDetail();
        return super.getDetail();
    }

    public XmlObject toXmlObject()
    {
        if ( m_description == null )
        {
            m_description = new Description[]{new Description( getFaultString() )};
        }
        XmlObject baseFaultXBean = createBaseFaultXmlBean();
        populateBaseFaultXmlBean( baseFaultXBean );
        return baseFaultXBean;
    }

    /**
     * Returns a string representation of this exception.
     *
     * @return a string representation of this exception
     */
    public String toString()
    {
        return new ToStringBuilder( this ).append( MSG.getMessage( Keys.FAULTCODE ),
                getFaultCode() )
                .append( MSG.getMessage( Keys.FAULTSTRING ),
                        getFaultString() ).append( MSG.getMessage( Keys.DESCRIPTION ), m_description )
                .append( MSG.getMessage( Keys.ERRORCODE ), m_errorCode ).append( MSG.getMessage( Keys.FAULTCAUSE ),
                        m_faultCause )
                .toString();
    }

    private XmlObject createBaseFaultXmlBean()
    {
        try
        {
            String docFactoryClassName = XmlBeanNameUtils.getDocumentElementXmlBeanFactoryClassName( getBaseFaultName() );
            Class docFactoryClass = Class.forName( docFactoryClassName );
            Method newInstanceMethod = docFactoryClass.getMethod( "newInstance", new Class[0] );
            XmlObject docXBean = (XmlObject) newInstanceMethod.invoke( null, new Object[0] );
            Class docXBeanClass = docXBean.getClass();
            Method addNewMethod = docXBeanClass.getMethod( "addNew" + XmlBeanNameUtils.getElementXmlBeanUnqualifiedClassName( getBaseFaultName() ), new Class[0] );
            XmlObject xBean = (XmlObject) addNewMethod.invoke( docXBean, new Object[0] );
            return xBean;
        }
        catch ( Exception e )
        {
            throw new JAXRPCException( "Failed to create strongly-typed XmlBean for global element with name " + getBaseFaultName() );
        }
    }

    private void populateBaseFaultXmlBean( XmlObject baseFaultXBean )
    {
        if ( m_namespaceSet instanceof org.apache.ws.resource.properties.v1_2_draft05.impl.NamespaceVersionHolderImpl )
        {
            org.oasisOpen.docs.wsrf.x2004.x11.wsrfWSBaseFaults12Draft03.BaseFaultType baseFaultType = (org.oasisOpen.docs.wsrf.x2004.x11.wsrfWSBaseFaults12Draft03.BaseFaultType) baseFaultXBean;
            baseFaultType.setTimestamp( m_timestamp );
            if ( m_description != null )
            {
                for ( int i = 0; i < m_description.length; i++ )
                {
                    org.oasisOpen.docs.wsrf.x2004.x11.wsrfWSBaseFaults12Draft03.BaseFaultType.Description description = baseFaultType.addNewDescription();
                    description.setStringValue( m_description[i].getValue() );
                    if ( m_description[i].getLang() != null )
                    {
                        description.setLang( m_description[i].getLang() );
                    }
                }
            }
            if ( m_errorCode != null )
            {
                org.oasisOpen.docs.wsrf.x2004.x11.wsrfWSBaseFaults12Draft03.BaseFaultType.ErrorCode errorCode = baseFaultType.addNewErrorCode();
                errorCode.setDialect( m_errorCode.getDialect().toString() );
                XmlObject[] elems = m_errorCode.getContent();
                for ( int i = 0; i < elems.length; i++ )
                {
                    XmlBeanUtils.addChildElement( errorCode, elems[i] );
                }
            }
            if ( m_faultCause != null )
            {
                for ( int i = 0; i < m_faultCause.length; i++ )
                {
                    XmlBeanUtils.addChildElement( baseFaultType, m_faultCause[i].toXmlObject() );
                }
            }
            if ( m_originator != null )
            {
                // TODO
            }
        }
        else
        {
            org.oasisOpen.docs.wsrf.x2004.x06.wsrfWSBaseFaults12Draft01.BaseFaultType baseFaultType = (org.oasisOpen.docs.wsrf.x2004.x06.wsrfWSBaseFaults12Draft01.BaseFaultType) baseFaultXBean;
            baseFaultType.setTimestamp( m_timestamp );
            if ( m_description != null )
            {
                for ( int i = 0; i < m_description.length; i++ )
                {
                    org.oasisOpen.docs.wsrf.x2004.x06.wsrfWSBaseFaults12Draft01.BaseFaultType.Description description = baseFaultType.addNewDescription();
                    description.setStringValue( m_description[i].getValue() );
                    if ( m_description[i].getLang() != null )
                    {
                        description.setLang( m_description[i].getLang() );
                    }
                }
            }
            if ( m_errorCode != null )
            {
                org.oasisOpen.docs.wsrf.x2004.x06.wsrfWSBaseFaults12Draft01.BaseFaultType.ErrorCode errorCode = baseFaultType.addNewErrorCode();
                errorCode.setDialect( m_errorCode.getDialect().toString() );
                XmlObject[] elems = m_errorCode.getContent();
                for ( int i = 0; i < elems.length; i++ )
                {
                    XmlBeanUtils.addChildElement( errorCode, elems[i] );
                }
            }
            if ( m_faultCause != null )
            {
                for ( int i = 0; i < m_faultCause.length; i++ )
                {
                    XmlBeanUtils.addChildElement( baseFaultType, m_faultCause[i].toXmlObject() );
                }
            }
            if ( m_originator != null )
            {
                // TODO
            }
        }
    }

    private void updateDetail()
    {
        XmlObject baseFaultDocXBean = toXmlObject();
        if ( m_detailEntry != null )
        {
            m_detailEntry.detachNode();
        }
        m_detailEntry = addDetailEntry( baseFaultDocXBean );
    }

}
