/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.impl;

import org.apache.commons.lang.SerializationException;
import org.apache.ws.resource.InvalidResourceKeyException;
import org.apache.ws.resource.ResourceKey;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPHeaderElement;

/**
 * LOG-DONE
 * A basic implementation of <code>ResourceKey</code>.
 */
public abstract class AbstractResourceKey
   implements ResourceKey
{
   private QName  m_name;
   private Object m_value;


   /**
    * Creates a new AbstractResourceKey.
    *
    * @param name  the name of the key
    * @param value the value of the key. The value of the key can be any simple/primitive type or any WSDL2Java
    *              generated type or any other type with proper type mappings.
    */
   public AbstractResourceKey( QName  name,
                               Object value )
   {
      if(name == null)
      {
          throw new IllegalArgumentException("Name may not be null!");
      }
      if(value == null)
      {
          throw new IllegalArgumentException("Value may not be null!");
      }
      m_name     = name;
      m_value    = value;
   }

   /**
    * Creates a new AbstractResourceKey.
    *
    * @param headerElem the SOAP header element containing the key. The element's name and namespace are used as key
    *                   name.
    * @param type       the type of the key. The SOAP header element will be deserialized into this type.
    *
    * @throws InvalidResourceKeyException if deserialization of the key fails.
    */
   public AbstractResourceKey( SOAPHeaderElement headerElem,
                               Class             type )
   throws InvalidResourceKeyException
   {
       if(headerElem == null)
       {
           throw new IllegalArgumentException("Header element may not be null!");
       }

           m_name =
              new QName( headerElem.getElementName(  ).getURI(  ),
                         headerElem.getElementName(  ).getLocalName(  ) );
           m_value = deserializeKey( headerElem, type );

   }

   /**
    * Creates a new AbstractResourceKey.
    *
    * @param headerElem the SOAPElement containing the key. The element's name and namespace are used as key name.
    *
    * @throws InvalidResourceKeyException if deserialization of the key fails.
    */
   public AbstractResourceKey( SOAPHeaderElement headerElem )
   throws InvalidResourceKeyException
   {
      this( headerElem, null );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public QName getName(  )
   {
      return m_name;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public Object getValue(  )
   {
      return m_value;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    *
    * @throws SerializationException DOCUMENT_ME
    */
   public abstract SOAPElement toSOAPElement(  )
   throws SerializationException;

   /**
    * DOCUMENT_ME
    *
    * @param obj DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public boolean equals( Object obj )
   {
      if ( this == obj )
      {
         return true;
      }

      if ( !( obj instanceof AbstractResourceKey ) )
      {
         return false;
      }

      AbstractResourceKey otherKey = (AbstractResourceKey) obj;

       //todo this will throw null pointer for singleton!
      return m_name.equals( otherKey.m_name ) && m_value.equals( otherKey.m_value );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public int hashCode(  )
   {    //todo this will throw null pointer for singleton!
      return m_name.hashCode(  ) + m_value.hashCode(  );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String toString(  )
   {      //todo this will throw null pointer for singleton!
      return m_name + "=" + m_value + " (type=" + m_value.getClass(  ).getName(  ) + ")";
   }

   /**
    * DOCUMENT_ME
    *
    * @param header DOCUMENT_ME
    * @param type   DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    *
    * @throws SerializationException      DOCUMENT_ME
    * @throws InvalidResourceKeyException DOCUMENT_ME
    */
   protected abstract Object deserializeKey( SOAPHeaderElement header,
                                             Class             type )
   throws SerializationException, 
          InvalidResourceKeyException;
}