/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.properties;

import org.apache.commons.lang.SerializationException;
import org.w3c.dom.Element;

import javax.xml.soap.SOAPElement;
import java.util.Iterator;

/**
 * Represents a single resource property. A resource property can have multiple values and can be converted into a DOM
 * <code>Element</code> or <code>SOAPElement</code> array.
 */
public interface ResourceProperty
{
    /**
     * Returns <tt>true</tt> if, and only if, this property contains no elements.
     *
     * @return <tt>true</tt> if, and only if, this property contains no elements
     */
    boolean isEmpty();

    /**
     * Gets meta data of this resource property.
     *
     * @return meta data of this resource property.
     */
    ResourcePropertyMetaData getMetaData();

    /**
     * Returns the resource property set that contains this property.
     *
     * @return the resource property set that contains this property.
     */
    ResourcePropertySet getSet();

    /**
     * Adds a value.
     *
     * @param value the value to add.
     *
     * @throws MetaDataViolationException if the name of the element to be added does not match the name associated with this property
     */
    void add( Object value );

    /**
     * Removes all values.
     */
    void clear();

    /**
     * Retrieves a value at a specific index.
     *
     * @param index the index of value to retrieve.
     *
     * @return the value at the given index. This operation might fail if the index is out of bounds.
     */
    Object get( int index );

    /**
     * Returns an iterator over the property elements.
     *
     * @return an iterator over the property elements
     */
    Iterator iterator();

    /**
     * Removes the first property element with the specified value.
     *
     * @param value value of a property element to remove
     *
     * @return true if, and only if, a property element was removed
     */
    boolean remove( Object value );

    /**
     * Sets a value at a specific index.
     *
     * @param index the index to set value at.
     * @param value the new value
     */
    void set( int index, Object value );

    /**
     * Returns the number of values in the resource property.
     *
     * @return the number of values.
     */
    int size();

    /**
     * Converts the resource property into an array of DOM {@link Element}s. If the RP has no values (is null), and RP
     * element was defined as: <ul> <li>minOccurs >= 0 - the function returns null.</li> <li>nillable == true - the
     * function returns a single element with <i>xsi:nil="true"</i> attribute set.</li> </ul>
     *
     * @return the resource property as a DOM Element array
     *
     * @throws SerializationException if conversion fails
     */
    Element[] toElements() throws SerializationException;

    /**
     * Converts the resource property value into an array of {@link SOAPElement}s. If the RP has no values (is null),
     * and RP element was defined as: <ul> <li>minOccurs >= 0 - the function returns null.</li> <li>nillable == true -
     * the function returns a single element with <i>xsi:nil="true"</i> attribute set.</li> </ul>
     *
     * @return the resource property as a SOAPElement array
     *
     * @throws SerializationException if conversion fails
     */
    SOAPElement[] toSOAPElements() throws SerializationException;

    /**
     * Returns the XML representation of this property (i.e. the resource property elements).
     *
     * @return the XML representation of this property
     */
    String toXML();

    /**
     * Sets a callback for this property.
     *
     * @param callback the callback
     */
    void setCallback( ResourcePropertyCallback callback );

    /**
     * Returns the callback for this property, or null if no callback has been set.
     *
     * @return the callback for this property, or null if no callback has been set
     */
    ResourcePropertyCallback getCallBack();

    ResourcePropertyValueChangeListener getChangeListener();

    void setChangeListener( ResourcePropertyValueChangeListener listenerValue );
}