/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.properties.query;

import org.apache.ws.resource.properties.ResourcePropertySet;
import org.w3c.dom.Element;

import java.net.URI;

/**
 * An <code>ExpressionEvaluator</code> is used to implement a query expression evaluation of a
 * <code>ResourcePropertySet</code>. An evaluator can be registered with a <code>QueryEngine</code>, which in turn calls
 * the evaluator when a matching expression is found.
 *
 * @see ResourcePropertySet
 * @see QueryEngine
 */
public interface ExpressionEvaluator
{
   /**
    * Gets the list of dialects that this evaluator can handle.
    *
    * @return list of dialects supported by this evaluator.
    */
   public URI[] getSupportedDialects(  );

   /**
    * Evaluates the expression over a ResourcePropertySet and returns the result.
    *
    * @param queryExpr           object passed by client representing query expression.
    * @param resourcePropertySet ResourcePropertySet associated with resource. The expression is evaluated against this
    *                            set.
    *
    * @return the result of the evaluation which depends on the expression. The results must be an xml serializable
    *         object in order to be passed back correctly to a remote client. The easiest way of achieving this is to
    *         model it as a Bean, or simply return a <code>SOAPElment</code> or <code>DOM Element</code>. If the result
    *         object returned is null an empty query result is returned.
    *
    * @throws UnknownQueryExpressionDialectException
    *                                       if query dialect is unsupported
    * @throws QueryEvaluationErrorException if query evaluation fails
    * @throws InvalidQueryExpressionException
    *                                       if query expression is invalid
    */
   public Object evaluate( QueryExpression     queryExpr,
                           ResourcePropertySet resourcePropertySet,
                           Element             nsContext )
   throws UnknownQueryExpressionDialectException, 
          QueryEvaluationErrorException, 
          InvalidQueryExpressionException;
}