/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.properties.query;

import org.apache.ws.resource.properties.ResourcePropertySet;
import org.w3c.dom.Element;

import java.net.URI;

/**
 * The <code>QueryEngine</code> interface is used to map queries on the service data of a service to the appropriate
 * <code>ExpressionEvaluators</code> and then return the result. <code>ExpressionEvaluators</code> can be preconfigured
 * or dynamically added at runtime.
 *
 * @see ExpressionEvaluator
 */
public interface QueryEngine
{
   /**
    * Gets the evaluator currently registered to handle an expression of the specified qualified name (from the top
    * level element of the XML Schema definition of the expression)
    *
    * @return the matching expression evaluator or null if none was found
    */
   public ExpressionEvaluator getEvaluator( URI dialect );

   /**
    * registers a new evaluator that can be used to evaluate queries on a service.
    *
    * @param evaluator implementation of evaluator to be used for evaluating queries specified in its XML Schema
    *                  definition.
    */
   public void registerEvaluator( ExpressionEvaluator evaluator );

   /**
    * DOCUMENT_ME
    *
    * @param dialect   DOCUMENT_ME
    * @param evaluator DOCUMENT_ME
    *
    * @throws UnknownQueryExpressionDialectException
    *          DOCUMENT_ME
    */
   public void registerEvaluator( URI                 dialect,
                                  ExpressionEvaluator evaluator )
   throws UnknownQueryExpressionDialectException;

   /**
    * Executes a query against a resource property set. The appropraite ExpressionEvaluator is used for the query based
    * on the dialect attribute.
    *
    * @param queryExpr       query expression
    * @param resourcePropSet resource properties set to execute the query against
    * @param nsContext       a DOM element to use as the context for resolving namespace prefixes
    *
    * @return the result of the evaluation which depends on the expression. The results must be an xml serializable
    *         object in order to be passed back correctly to a remote client. The easiest way of achieving this is to
    *         model it as a Bean, or simply return a <code>SOAPElment</code> or <code>DOM Element</code>. If the result
    *         object returned is null an empty query result is returned.
    *
    * @throws UnknownQueryExpressionDialectException
    *                                       if query dialect is unsupported
    * @throws QueryEvaluationErrorException if query evaluation fails
    * @throws InvalidQueryExpressionException
    *                                       if query expression is invalid
    */
   Object executeQuery( QueryExpression     queryExpr,
                        ResourcePropertySet resourcePropSet,
                        Element             nsContext )
   throws UnknownQueryExpressionDialectException, 
          QueryEvaluationErrorException, 
          InvalidQueryExpressionException;
}