/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.tool;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.runtime.RuntimeConstants;
import org.apache.ws.resource.ResourceDefinition;
import org.apache.ws.resource.i18n.Keys;
import org.apache.ws.resource.i18n.MessagesImpl;
import org.apache.ws.resource.impl.ResourceDefinitionImpl;
import org.apache.ws.resource.tool.velocity.ImplementsListBuilder;
import org.apache.ws.resource.tool.velocity.ServiceProperties;
import org.apache.ws.util.XmlBeanNameUtils;
import org.apache.ws.util.i18n.Messages;
import org.apache.ws.util.velocity.CommonsLogLogSystem;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.impl.common.XmlErrorPrinter;
import org.apache.xmlbeans.impl.tool.SchemaCompiler;

import javax.wsdl.Definition;
import javax.wsdl.Service;
import javax.wsdl.Port;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * Generates Java Web service types and skeletons from a WSRF WSDL.
 *
 * @author Ian Springer (ian DOT springer AT hp DOT com)
 */
public class Wsdl2Java
{
    private static final Messages MSG = MessagesImpl.getInstance();

    /**
     * DOCUMENT_ME
     */
    protected static final Options CMD_LINE_OPTIONS = new Options();

    static
    {
        initCmdLineOptions();
    }

    private File[] m_wsdlFiles;
    private Wsdl2JavaOptions m_options;
    private Map m_portTypeInfoMap = new HashMap();
    private WSDLReader m_wsdlReader;
    private File m_outputDir;
    private File[] m_classpath;
    private File m_xmlbeansDir;

    /**
     * Creates a new {@link Wsdl2Java} object.
     *
     * @param wsdlFiles DOCUMENT_ME
     * @param outputDir DOCUMENT_ME
     * @param classpathFiles
     * @param options   DOCUMENT_ME
     */
    public Wsdl2Java( File[] wsdlFiles,
                      File outputDir,
                      File[] classpathFiles,
                      Wsdl2JavaOptions options ) throws Exception
    {
        m_wsdlFiles = wsdlFiles;
        if ( m_wsdlFiles.length == 0 )
        {
            System.out.println( MSG.getMessage( Keys.WARN_EMPTY_WSDLS ) );
        }
        m_outputDir = outputDir;
        m_xmlbeansDir = new File( m_outputDir, ".xmlbeans" );
        m_classpath = classpathFiles;
        validateClasspath();
        m_options = ( options != null ) ? options : new Wsdl2JavaOptions();
        m_wsdlReader = WSDLFactory.newInstance().newWSDLReader();
        initVelocity();
        initPortTypeInfoMap();
    }

    private void validateClasspath()
            throws MalformedURLException
    {
        URLClassLoader cpClassLoader = new URLClassLoader( toURLs( m_classpath ), null );
        try
        {
            cpClassLoader.loadClass( XmlObject.class.getName() );
        }
        catch ( ClassNotFoundException cnfe )
        {
            throw new IllegalStateException( "Specified classpath must contain the XMLBeans runtime jar (e.g. xbean-v1HEAD-SNAPSHOT-20041129.jar)" );
        }
        try
        {
            cpClassLoader.loadClass( org.oasisOpen.docs.wsrf.x2004.x06.wsrfWSResourceProperties12Draft01.GetResourcePropertyDocument.class.getName() );
        }
        catch ( ClassNotFoundException cnfe )
        {
            throw new IllegalStateException( "Specified classpath must contain the WSRF generated XMLBeans jar (e.g. wsrf-xbeans-SNAPSHOT.jar)" );
        }
    }

    private URL[] toURLs( File[] files ) throws MalformedURLException
    {
        URL[] urls = new URL[files.length];
        for ( int i = 0; i < files.length; i++ )
        {
            urls[i] = files[i].toURL();
        }
        return urls;
    }

    /**
     * DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    public Wsdl2JavaOptions getOptions()
    {
        return m_options;
    }

    /**
     * DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    public File getOutputDir()
    {
        return m_outputDir;
    }

    /**
     * DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    public File[] getWsdlFiles()
    {
        return m_wsdlFiles;
    }

    /**
     * DOCUMENT_ME
     *
     * @throws Exception DOCUMENT_ME
     */
    public void generate()
            throws Exception
    {
        for ( int i = 0; i < m_wsdlFiles.length; i++ )
        {
            processWsdlFile( m_wsdlFiles[i] );
        }
    }

    private void processWsdlFile( File wsdlFile )
            throws Exception
    {
        System.out.println( "Processing WSDL file \"" + wsdlFile + "\"..." );
        Definition def = m_wsdlReader.readWSDL( wsdlFile.getPath() );
        Iterator serviceIter = def.getServices().values().iterator();
        while ( serviceIter.hasNext() )
        {
            Service service = (Service) serviceIter.next();
            Iterator iterator = service.getPorts().values().iterator();
            while (iterator.hasNext())
            {
                Port port = (Port) iterator.next();
                ResourceDefinition resourceDef = new ResourceDefinitionImpl( def, port, wsdlFile.toURL());
                File serviceDir = new File( m_outputDir, resourceDef.getName() );
                System.out.println( "Processing WSRF WSDL port \"" + resourceDef.getName() + "\"..." );
                processTemplates( resourceDef, wsdlFile, serviceDir );
                copyWsdlFile( wsdlFile, serviceDir );
            }
        }
        generateXmlBeans( wsdlFile );
    }

    /**
     * DOCUMENT_ME
     *
     * @param args DOCUMENT_ME
     *
     * @throws ParseException DOCUMENT_ME
     */
    public static void main( String[] args )
            throws ParseException
    {
        CommandLine cmdLine = new PosixParser().parse( CMD_LINE_OPTIONS, args, true );
        checkForRequiredOption( cmdLine, Opts.OUTPUT_DIR );
        checkForRequiredOption( cmdLine, Opts.CLASSPATH );
        File outputDir = new File( cmdLine.getOptionValue( Opts.OUTPUT_DIR ) );
        File[] classpathFiles = toFileArray( cmdLine.getOptionValue( Opts.CLASSPATH ) );

        Wsdl2JavaOptions options = new Wsdl2Java.Wsdl2JavaOptions();

        if ( cmdLine.hasOption( Opts.VERBOSE ) )
        {
            options.setVerbose( Boolean.valueOf( cmdLine.getOptionValue( Opts.VERBOSE ) ).booleanValue() );
        }

        if ( cmdLine.hasOption( Opts.DEBUG ) )
        {
            options.setVerbose( Boolean.valueOf( cmdLine.getOptionValue( Opts.DEBUG ) ).booleanValue() );
        }

        File[] wsdlFiles = new File[cmdLine.getArgs().length];
        for ( int i = 0; i < cmdLine.getArgs().length; i++ )
        {
            wsdlFiles[i] = new File( cmdLine.getArgs()[i] );
        }

        try
        {
            System.out.println( MSG.getMessage( Keys.WSDL4J_PASSED_ARGUMENTS,
                    Integer.toString( wsdlFiles.length ),
                    outputDir ) );
            new Wsdl2Java( wsdlFiles, outputDir, classpathFiles, options ).generate();
        }
        catch ( Exception e )
        {
            e.printStackTrace();
            System.exit( 1 );
        }
    }

    /**
     * DOCUMENT_ME
     *
     * @param ptName DOCUMENT_ME
     *
     * @return DOCUMENT_ME
     */
    protected PortType2JavaInfo getPortType2JavaInfo( QName ptName )
    {
        return (PortType2JavaInfo) m_portTypeInfoMap.get( ptName );
    }

    /**
     * DOCUMENT_ME
     *
     * @param ptInfo DOCUMENT_ME
     */
    protected void addPortType2JavaInfo( PortType2JavaInfo ptInfo )
    {
        m_portTypeInfoMap.put( ptInfo.getName(),
                ptInfo );
    }

    private static void checkForRequiredOption( CommandLine cmdLine,
                                                String opt )
    {
        if ( !cmdLine.hasOption( opt ) )
        {
            System.err.println( MSG.getMessage( Keys.OPT_REQUIRED, opt ) );
            System.exit( 1 );
        }
    }

    private static void initCmdLineOptions()
    {
        CMD_LINE_OPTIONS.addOption( Opts.OUTPUT_DIR,
                LongOpts.OUTPUT_DIR,
                true,
                MSG.getMessage( Keys.OPT_OUTPUT_DIR_FOR_GEN_SRC ) );
        CMD_LINE_OPTIONS.addOption( Opts.CLASSPATH,
                LongOpts.CLASSPATH,
                true,
                MSG.getMessage( Keys.OPT_CLASSPATH_SENT_TO_XMLBEANS ) );
        CMD_LINE_OPTIONS.addOption( Opts.VERBOSE,
                LongOpts.VERBOSE,
                false,
                MSG.getMessage( Keys.OPT_ENABLE_VERBOSE ) );
        CMD_LINE_OPTIONS.addOption( Opts.DEBUG,
                LongOpts.DEBUG,
                false,
                MSG.getMessage( Keys.OPT_ENABLE_DEBUG ) );
    }

    private String getJarFileName( File wsdlFile )
    {
        String wsdlFileName = wsdlFile.getName();
        String baseName = wsdlFileName.substring( 0, wsdlFileName.lastIndexOf( "." ) );
        return baseName + "-xbeans.jar";
    }

    private File getPackageDir( String targetNamespace, File serviceDir )
    {
        String javaPackageName = GenerationUtils.getJavaPackageName( targetNamespace );
        File javaDir = new File( serviceDir, "src/java" );
        return new File( javaDir,
                javaPackageName.replace( '.', '/' ) );
    }

    private File getTempDir()
    {
        String userTempDir = System.getProperty( "java.io.tmpdir" );
        File userDir = new File( userTempDir );
        userDir.mkdirs();
        File tempClassesDir = new File( userDir, "tmp" + new Date().getTime() );
        tempClassesDir.mkdir();
        return tempClassesDir;
    }

    private ServiceProperties buildServiceProperties( ResourceDefinition resourceDef,
                                                      File wsdlFile,
                                                      File serviceDir )
    {
        ServiceProperties props = new ServiceProperties(resourceDef);        
        String targetNamespace = resourceDef.getDefinition().getTargetNamespace();
        props.setPackageName( GenerationUtils.getJavaPackageName( targetNamespace ) );
        String serviceName = resourceDef.getName();
        String capitalizedServiceName = StringUtils.capitalize( serviceName );
        props.setServiceName( capitalizedServiceName );
        props.setSrcDir( new File( serviceDir, "src/java" ).getAbsolutePath() );
        props.setJarDir( new File(m_xmlbeansDir,"lib").getAbsolutePath() );
        props.setJarFileName( getJarFileName( wsdlFile ) );
        props.setClassesDir( new File( serviceDir, "classes" ).getAbsolutePath() );
        props.setEndpointURL( resourceDef.getEndpointURL() );
        props.setEndpointServiceName( serviceName );
        props.setServiceImplements( ImplementsListBuilder.getServiceImplementsList( resourceDef, m_portTypeInfoMap ) );
        props.setNamespace( targetNamespace );
        props.setPrefix( GenerationUtils.getPrefix( targetNamespace ) );
        props.setCustomMappings( GenerationUtils.getOperationInfoMap( resourceDef ) );
        props.setHasProperties( resourceDef.hasProperties() );
        QName[] customPropertyNames = resourceDef.getCustomPropertyNames();
        Map customPropMap = new HashMap();
        for ( int i = 0; i < customPropertyNames.length; i++ )
        {
            QName customPropertyName = customPropertyNames[i];
            String propName = customPropertyName.getLocalPart().toUpperCase();
            int count = 2;
            while ( customPropMap.containsKey( propName ) )
            {
                propName = propName + count;
                count++;
            }

            customPropMap.put( propName, customPropertyName );
        }
        props.setPropertyMap( customPropMap );
        props.setResourceImplements( ImplementsListBuilder.getResourceImplementsList( resourceDef, m_portTypeInfoMap ) );
        QName propertiesDocumentName = resourceDef.getPropertiesDocumentName();
        if( propertiesDocumentName != null)
        {
        props.setPropertyDocumentName(
                XmlBeanNameUtils.getDocumentElementXmlBeanClassName(
                  propertiesDocumentName       ) );
        }
        props.setWsdlName( wsdlFile.getName() );
        props.setResourcekey( new QName( targetNamespace, "ResourceID" ) );
        Iterator ptNameIter = resourceDef.getImplementedResourceCapabilities().keySet().iterator();
        while ( ptNameIter.hasNext() )
        {
            QName ptName = (QName) ptNameIter.next();
            PortType2JavaInfo ptInfo = (PortType2JavaInfo) m_portTypeInfoMap.get( ptName );
            if ( ptInfo != null )
            {
                if ( ptInfo.getResourceTemplateFileName() != null )
                {
                    props.addResourceIncludeFile( ptInfo.getResourceTemplateFileName() );
                }
                if ( ptInfo.getResourceInitMethodIncludeLocation() != null )
                {
                    props.addResourceInitIncludeLocation( ptInfo.getResourceInitMethodIncludeLocation() );
                }
                if ( ptInfo.getAbstractResourceInitMethodIncludeLocation() != null )
                {
                    props.addAbstractResourceInitIncludeLocation( ptInfo.getAbstractResourceInitMethodIncludeLocation() );
                }
                if ( ptInfo.getServiceTemplateFileName() != null )
                {
                    props.addServiceIncludeFile( ptInfo.getServiceTemplateFileName() );
                }
            }
        }
        return props;
    }

    private void generateXmlBeans( File wsdlFile )
            throws Exception
    {
        System.out.println( "Generating XMLBeans for WSDL file \"" + wsdlFile + "\"..." );
        SchemaCompiler.Parameters scompParams = new SchemaCompiler.Parameters();
        scompParams.setSrcDir( new File( m_xmlbeansDir, "src" ) );
        File tempClassesDir = getTempDir();
        scompParams.setClassesDir( tempClassesDir );
        scompParams.setWsdlFiles( new File[]
        {
            wsdlFile
        } );
        scompParams.setDownload( true );
        scompParams.setClasspath( m_classpath );
        scompParams.setVerbose( m_options.isVerbose() );
        scompParams.setDebug( m_options.isDebug() );
        final boolean beNoisy = true;
        XmlErrorPrinter xErrorListener = new XmlErrorPrinter( beNoisy, null );
        scompParams.setErrorListener( xErrorListener );
        File jarDir = new File( m_xmlbeansDir, "lib" );
        jarDir.mkdirs(  );
        String wsdlJarFileName = getJarFileName( wsdlFile );
        scompParams.setOutputJar( new File( jarDir, wsdlJarFileName ) );
        try
        {
            if ( !SchemaCompiler.compile( scompParams ) )
            {
                throw new Exception( xErrorListener.toString() );
            }
        }
        finally
        {
            //delete the temp directory
            FileUtils.deleteDirectory( tempClassesDir );
        }
    }

    private void copyWsdlFile( File wsdlFile, File serviceDir )
            throws IOException
    {
        File javaDir = new File( serviceDir, "src" );
        File wsdlDir = new File( javaDir, "wsdl" );
        wsdlDir.mkdirs();
        FileUtils.copyFile( wsdlFile, new File( wsdlDir, wsdlFile.getName() ) );
    }

    private void initPortTypeInfoMap()
    {
        //draft 01
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft01.GetResourcePropertyPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft01.GetMultipleResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft01.SetResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft01.QueryResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft01.ImmediateResourceTerminationPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft01.ScheduledResourceTerminationPortType2JavaInfo() );

        //draft 05
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.GetResourcePropertyPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.GetMultipleResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.SetResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.QueryResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.ImmediateResourceTerminationPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.ScheduledResourceTerminationPortType2JavaInfo() );

        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.InsertResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.UpdateResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.DeleteResourcePropertiesPortType2JavaInfo() );
        addPortType2JavaInfo( new org.apache.ws.resource.tool.porttype.v1_2_draft05.GetResourcePropertyDocumentPortType2JavaInfo() );
    }

    private void initVelocity()
            throws Exception
    {
        // configure to use Commons Logging for logging
        Velocity.setProperty( RuntimeConstants.RUNTIME_LOG_LOGSYSTEM_CLASS, CommonsLogLogSystem.class.getName() );
        // configure to use classpath-based resource loader
        Velocity.addProperty( RuntimeConstants.RESOURCE_LOADER, "classpath" );
        String resourceLoaderBaseKey = "classpath." + RuntimeConstants.RESOURCE_LOADER + ".";
        Velocity.setProperty( resourceLoaderBaseKey + "class", "org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader" );
        Velocity.setProperty( resourceLoaderBaseKey + "cache", "false" );
        Velocity.setProperty( resourceLoaderBaseKey + "modificationCheckInterval", "2" );
        Velocity.init();
    }

    private void processTemplate( VelocityContext context,
                                  String templateLocation,
                                  File outputFile )
            throws Exception
    {
        /*
         *  get the Template object.  This is the parsed version of your
         *  template input file.  Note that getTemplate() can throw
         *   ResourceNotFoundException : if it doesn't find the template
         *   ParseErrorException : if there is something wrong with the VTL
         *   Exception : if something else goes wrong (this is generally
         *        indicative of as serious problem...)
         */
        try
        {
            Template template = Velocity.getTemplate( templateLocation );
            /*
             *  Now have the template engine process your template using the
             *  data placed into the context. Think of it as a  'merge'
             *  of the template and the data to produce the output stream.
             */
            FileWriter writer = new FileWriter( outputFile );
            if ( template != null )
            {
                template.merge( context, writer );
            }
            writer.close();
        }
        catch ( Exception e )
        {
            System.err.println( "Error processing template " + templateLocation );
            e.printStackTrace();
        }
    }

    private void processTemplates( ResourceDefinition resourceDef,
                                   File wsdlFile, File serviceDir )
            throws Exception
    {
        String capitalizedServiceName = StringUtils.capitalize( resourceDef.getName() );
        try
        {
            VelocityContext context = new VelocityContext();
            ServiceProperties props = buildServiceProperties( resourceDef, wsdlFile, serviceDir );

            //add to context
            context.put( "generated", props );

            //generate files
            File packageDir = getPackageDir( resourceDef.getDefinition().getTargetNamespace(), serviceDir );
            packageDir.mkdirs();
            File outputFile = new File( packageDir, "Abstract" + capitalizedServiceName + "Service.java" );
            processTemplate( context, "templates/AbstractService.vm", outputFile );

            outputFile = new File( packageDir, capitalizedServiceName + "Service.java" );

            //only generate if it doesn't exist
            if ( !outputFile.exists() )
            {
                processTemplate( context, "templates/Service.vm", outputFile );
            }

            outputFile = new File( packageDir, "Abstract" + capitalizedServiceName + "Resource.java" );
            processTemplate( context, "templates/AbstractResource.vm", outputFile );

            outputFile = new File( packageDir, capitalizedServiceName + "Resource.java" );

            //only generate if it doesn't exist
            if ( !outputFile.exists() )
            {
                processTemplate( context, "templates/Resource.vm", outputFile );
            }

            outputFile = new File( packageDir, capitalizedServiceName + "Home.java" );

            //only generate if it doesn't exist
            if ( !outputFile.exists() )
            {
                processTemplate( context, "templates/Home.vm", outputFile );
            }

            //only if they implemented properties does this make sense.
            if ( resourceDef.hasProperties() )
            {
                outputFile = new File( packageDir, capitalizedServiceName + "PropertyQNames.java" );
                processTemplate( context, "templates/PropertyQNames.vm", outputFile );
            }

            outputFile = new File( packageDir, capitalizedServiceName + "_deploy.wsdd" );
            processTemplate( context, "templates/deploy.vm", outputFile );

            outputFile = new File( packageDir, capitalizedServiceName + "_jndi-config.xml" );
            processTemplate( context, "templates/jndi.vm", outputFile );

            outputFile = new File( packageDir, capitalizedServiceName + "CustomOperationsPortType.java" );
            processTemplate( context, "templates/CustomOperationsPortType.vm", outputFile );

            outputFile = new File( serviceDir, "build.xml" );
            processTemplate( context, "templates/build.xml.vm", outputFile );

            outputFile = new File( serviceDir, "build.properties" );
            processTemplate( context, "templates/build.properties.vm", outputFile );
        }
        catch ( Exception e )
        {
            e.printStackTrace();
        }
    }

    /**
     * DOCUMENT_ME
     *
     * @author $author$
     * @version $Revision: 1.8 $
     */
    static interface LongOpts
    {
        /**
         * DOCUMENT_ME
         */
        String OUTPUT_DIR = "outputDir";

        /**
         * DOCUMENT_ME
         */
        String VERBOSE = "verbose";

        /**
         * DOCUMENT_ME
         */
        String DEBUG = "debug";

        /**
         * DOCUMENT_ME
         */
        String CLASSPATH = "classpath";
    }

    /**
     * DOCUMENT_ME
     *
     * @author $author$
     * @version $Revision: 1.8 $
     */
    static interface Opts
    {
        /**
         * DOCUMENT_ME
         */
        String OUTPUT_DIR = "o";

        /**
         * DOCUMENT_ME
         */
        String VERBOSE = "v";

        /**
         * DOCUMENT_ME
         */
        String DEBUG = "d";

        /**
         * DOCUMENT_ME
         */
        String CLASSPATH = "C";
    }

    /**
     * DOCUMENT_ME
     *
     * @author $author$
     * @version $Revision: 1.8 $
     */
    public static class Wsdl2JavaOptions
    {
        private boolean m_verbose;
        private boolean m_debug;

        /**
         * DOCUMENT_ME
         *
         * @param debug DOCUMENT_ME
         */
        public void setDebug( boolean debug )
        {
            m_debug = debug;
        }

        /**
         * DOCUMENT_ME
         *
         * @return DOCUMENT_ME
         */
        public boolean isDebug()
        {
            return m_debug;
        }

        /**
         * DOCUMENT_ME
         *
         * @param verbose DOCUMENT_ME
         */
        public void setVerbose( boolean verbose )
        {
            m_verbose = verbose;
        }

        /**
         * DOCUMENT_ME
         *
         * @return DOCUMENT_ME
         */
        public boolean isVerbose()
        {
            return m_verbose;
        }
    }

    static File[] toFileArray( String classpath )
    {
        List classpathItems = new ArrayList();
        for ( StringTokenizer tokenizer = new StringTokenizer( classpath, File.pathSeparator );
              tokenizer.hasMoreTokens(); )
        {
            classpathItems.add( new File( tokenizer.nextToken() ) );
        }
        return (File[]) classpathItems.toArray( new File[0] );
    }

}