/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.util.i18n.Keys;
import org.apache.ws.util.i18n.Messages;
import org.apache.ws.util.i18n.MessagesImpl;
import sun.io.Converters;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

/**
 * LOG-DONE
 * Generic I/O utility methods.
 *
 * @author Ian P. Springer, Sal Campana
 */
public abstract class IoUtils
{

   private static final Log LOG = LogFactory.getLog( IoUtils.class );
   private static final Messages MSG = MessagesImpl.getInstance();

   /**
    * Copies an input stream to an output stream.
    *
    * @param in  an input stream; must not be null
    * @param out an output stream; must not be null
    *
    * @throws IOException if the copy fails
    */
   public static void copy( InputStream  in,
                            OutputStream out )
   throws IOException
   {
      if ( in == null )
      {
         throw new IllegalArgumentException( MSG.getMessage( Keys.INPUTSTRM_MST_NOT_B_NULL) );
      }
      if ( out == null )
      {
         throw new IllegalArgumentException( MSG.getMessage( Keys.OUTSTRM_MST_NOT_B_NULL) );
      }
      // use Buffered streams for better performance
      in     = new BufferedInputStream( in );
      out    = new BufferedOutputStream( out );

      // do not allow other threads to read from the input or write to
      // the output while copying is taking place
      synchronized ( in )
      {
         synchronized ( out )
         {
            byte[] bytes     = new byte[8192];
            int    bytesRead;
            while ( ( bytesRead = in.read( bytes ) ) != -1 )
            {
               out.write( bytes, 0, bytesRead );
            }

            out.flush(  );
         }
      }
   }

   /**
       * Copies an input stream to a file.
       *
       * @param in an input stream to copy from
       * @param targetFile the file to copy to; if the file already exists, it will be overwritten
       *
       * @throws IOException if the copy fails
       */
      public static void copy( InputStream in,
                               File targetFile )
      throws IOException
      {
         LOG.debug(MSG.getMessage( Keys.COPY_OUTPUTSTRM_TO_FILE, targetFile.toString()));
         File targetDir = targetFile.getAbsoluteFile(  ).getParentFile(  );
         targetDir.mkdirs(  );
         File tempFile = File.createTempFile( MSG.getMessage(Keys.COPY)+"-" + targetFile.getName(  ) + "-", "", targetDir );
         try
         {
            OutputStream out = new FileOutputStream( tempFile );
            copy( in, out );
            in.close(  );
            out.close(  );
            if ( targetFile.exists(  ) && targetFile.canWrite(  ) )
            {
               targetFile.delete(  );
            }

            if ( !tempFile.renameTo( targetFile ) )
            {
               throw new IOException( MSG.getMessage( Keys.FAILED_TO_COPY_INPUTSTRM,targetFile.getAbsolutePath(  ) ,
                                       tempFile.getAbsolutePath(  )) );
            }
         }
         catch ( IOException ioe )
         {
            tempFile.delete(  );
            throw ( ioe );
         }
      }

   /**
    * Copies one file to another.
    *
    * @param sourceFile an existing file to copy from
    * @param targetFile the file to copy to; if the file already exists, it will be overwritten
    *
    * @throws IOException if the copy fails
    */
   public static void copy( File sourceFile,
                            File targetFile )
   throws IOException
   {
      LOG.debug(MSG.getMessage( Keys.ATTMPT_COPY_FILE_TO_FILE,sourceFile,targetFile));
      try
      {
         copy( new FileInputStream( sourceFile ), targetFile );
      }
      catch ( IOException ioe )
      {
         throw new IOException( MSG.getMessage( Keys.FAILED_TO_CPY_FILE_TO_FILE,sourceFile.getAbsolutePath(  )
                                   , targetFile.getAbsolutePath(  ) , ioe.getMessage() ));
      }
   }

   /**
    * Deletes all files and subdirectories under dir. Returns true if all deletions were successful. If a deletion
    * fails, the method stops attempting to delete and returns false.
    */
   public static boolean deleteDir( File dir )
   {
      LOG.debug(MSG.getMessage( Keys.ATTMPT_DEL_DIR, dir.toString()));
      if ( dir.isDirectory(  ) )
      {
         String[] children = dir.list(  );
         for ( int i = 0; i < children.length; i++ )
         {
            boolean success = deleteDir( new File( dir, children[i] ) );
            if ( !success )
            {
               return false;
            }
         }
      }

      // The directory is now empty so delete it
      return dir.delete(  );
   }

   /**
    * Converts a ByteArrayOutputStream into a ByteArrayInputStream.
    *
    * @param out
    *
    * @return ByteArrayInputStream
    */
   public static ByteArrayInputStream toByteArrayInputStream( ByteArrayOutputStream out )
   {
      return new ByteArrayInputStream( out.toByteArray(  ) );
   }

   /**
    * Convert the specified string to an input stream.
    *
    * @param str a string
    *
    * @return an input stream
    */
   public static InputStream toInputStream( String str )
   {
      LOG.debug(MSG.getMessage( Keys.ATTMPT_CNVRT_STRING_TO_INPTSTRM, str));
      try
      {
         return ( new ByteArrayInputStream( str.getBytes( "UTF-8" ) ) );
      }
      catch ( UnsupportedEncodingException uee )
      { // this should never occur - UTF-8 is always supported
         uee.printStackTrace(  );

         // return using the current locale encoding
         return ( new ByteArrayInputStream( str.getBytes(  ) ) );
      }
   }

   /**
    * TODO
    *
    * @param in
    *
    * @return
    *
    * @throws IOException
    */
   public static String toString( InputStream in )
   throws IOException
   {
      return toString( in,
                       Converters.getDefaultEncodingName(  ) );
   }

   /**
    * TODO
    *
    * @param in
    * @param encoding
    *
    * @return
    *
    * @throws IOException
    */
   public static String toString( InputStream in,
                                  String      encoding )
   throws IOException
   {
      ByteArrayOutputStream out = new ByteArrayOutputStream(  );
      copy( in, out );
      return out.toString( encoding );
   }

}
