/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.util.i18n.Keys;
import org.apache.ws.util.i18n.Messages;
import org.apache.ws.util.i18n.MessagesImpl;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

/**
 * A set of generic JAXP utility methods.
 *
 * @author Ian P. Springer
 */
public abstract class JaxpUtils
{
    private static final Log LOG = LogFactory.getLog( JaxpUtils.class );
    public static final Messages MSG = MessagesImpl.getInstance();

    /**
     * DOCUMENT_ME
     */
    private static final Object TRANSFORMER_LOCK = new Object();

    /**
     * DOCUMENT_ME
     */
    private static final String INDENT_SPACES = "2";

    /**
     * thread local that gives each thread its own TransformerFactory (since it is not thread-safe)
     */
    public static ThreadLocal TRANSFORMER_FACTORIES =
            new ThreadLocal()
            {
                protected synchronized Object initialValue()
                {
                    return TransformerFactory.newInstance();
                }
            };

    /**
     * thread local that gives each thread its own DocumentBuilderFactory (since it is not thread-safe)
     */
    public static ThreadLocal DOCUMENT_BUILDER_FACTORIES =
            new ThreadLocal()
            {
                protected synchronized Object initialValue()
                {
                    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                    dbf.setNamespaceAware( true );

                    return ( dbf );
                }
            };

    /**
     * Loads a DOM Document from an InputSource.
     *
     * @param in_source
     *
     * @return Document
     *
     * @throws ParserConfigurationException
     * @throws SAXException
     * @throws IOException
     */
    public static Document loadDocument( InputSource in_source )
            throws ParserConfigurationException,
            SAXException,
            IOException
    {
        DocumentBuilderFactory dbf = (DocumentBuilderFactory) DOCUMENT_BUILDER_FACTORIES.get();
        DocumentBuilder db = dbf.newDocumentBuilder();

        return ( db.parse( in_source ) );
    }

    /**
     * Method to load a DOM Document from an InputStream.
     *
     * @param in_stream
     *
     * @return Document
     *
     * @throws ParserConfigurationException
     * @throws SAXException
     * @throws IOException
     */
    public static Document loadDocument( InputStream in_stream )
            throws ParserConfigurationException,
            SAXException,
            IOException
    {
        InputStreamReader utf8_stream = new InputStreamReader( in_stream, "UTF-8" );
        Document doc = loadDocument( new InputSource( utf8_stream ) );
        utf8_stream.close();

        return ( doc );
    }

    /**
     * Converts a string containing an XML document into a {@link Document}.
     *
     * @param xmlString
     *
     * @return the W3C DOM document
     *
     * @throws ParserConfigurationException
     * @throws SAXException
     * @throws IOException
     */
    public static Document toDocument( String xmlString )
            throws ParserConfigurationException,
            SAXException,
            IOException
    {
        LOG.debug( MSG.getMessage( Keys.LOAD_XML_STRING, xmlString ) );
        return ( loadDocument( IoUtils.toInputStream( xmlString ) ) );
    }

    /**
     * Converts an XML {@link Node} into a string containing the corresponding XML. ***NOTE*** requires an XSLT
     * implementation (ie - Xalan) in the classpath at runtime
     *
     * @param node a DOM node
     *
     * @return String representation of the node
     *
     * @throws Exception
     */
    public static String toString( Node node )
            throws Exception
    {
        return ( toString( node, false ) );
    }

    /**
     * Converts an XML {@link Node} into a string containing the corresponding XML, optionally pretty-printed with
     * indenting and newlines. ***NOTE*** requires an XSLT implementation (ie - Xalan) in the classpath at runtime
     *
     * @param node         a DOM node
     * @param indentOutput whether or not to indent the output XML
     *
     * @return String representation of the node
     *
     * @throws Exception
     */
    public static String toString( Node node,
                                   boolean indentOutput )
            throws Exception
    {
        /**
         * Due to a bug in the Transformer implementation we need to
         * synchronize access to the transform() method.
         */
        synchronized ( TRANSFORMER_LOCK )
        {
            TransformerFactory transformer_factory = (TransformerFactory) TRANSFORMER_FACTORIES.get();
            Transformer transformer = transformer_factory.newTransformer();

            if ( indentOutput )
            {
                transformer.setOutputProperty( TransformConstants.OUTPUT_PROP_INDENT, "yes" );
                transformer.setOutputProperty( TransformConstants.OUTPUT_PROP_INDENT_AMOUNT, INDENT_SPACES );
            }

            DOMSource dom_source = new DOMSource( node );
            ByteArrayOutputStream doc_stream = new ByteArrayOutputStream();
            StreamResult result = new StreamResult( doc_stream );

            transformer.transform( dom_source, result );

            return ( doc_stream.toString() );
        }
    }
}