/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.util.i18n.Keys;
import org.apache.ws.util.i18n.Messages;
import org.apache.ws.util.i18n.MessagesImpl;

import javax.xml.namespace.QName;
import javax.xml.soap.Name;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFactory;
import java.util.Hashtable;
import java.util.Map;

/**
 * @author Ian P. Springer (Hewlett-Packard Company), Sal Campana (Hewlett-Packard Company)
 */
public abstract class NameUtils
{
   private static final Log LOG = LogFactory.getLog( NameUtils.class );
    public static final Messages MSG = MessagesImpl.getInstance();

   private static Map s_namespace_to_prefix_map = new Hashtable(  );

   /**
    * Builds a valid concise prefix for the specified namespace URI.
    *
    * @param nsURI the namespace URI
    *
    * @return the prefix
    */
   public static String getNamespacePrefix( String nsURI )
   {
      LOG.debug(MSG.getMessage( Keys.BLD_PREFIX_FOR_NAMESPC,nsURI));
      if ( nsURI == null )
      {
         return "";
      }

      //first check to see if we've done this yet
      if ( s_namespace_to_prefix_map.containsKey( nsURI ) )
      {
         return (String) s_namespace_to_prefix_map.get( nsURI );
      }

      // haven't done it yet so do it now
      String prefix = nsURI;

      while ( ( prefix.length(  ) > 0 ) && !Character.isLetterOrDigit( prefix.charAt( prefix.length(  ) - 1 ) ) )
      {
         prefix = prefix.substring( 0, prefix.length(  ) - 1 );
      }

      int start_index = prefix.lastIndexOf( '/' );

      if ( start_index == -1 )
      {
         start_index = prefix.lastIndexOf( '.' );
      }

      if ( start_index == -1 )
      {
         start_index = prefix.lastIndexOf( ':' );
      }

      //todo commented until better solution is completed
      //boolean vs Boolean....this would generate the same prefix as it stands....
      //prefix = prefix.substring( start_index + 1 ).toLowerCase();
      prefix = prefix.substring( start_index + 1 );

      String newPrefix = prefix;

      for ( int i = 0; i < prefix.length(  ); i++ )
      {
         char current_char = prefix.charAt( i );

         if ( !isValidPrefixChar( current_char ) )
         {
            newPrefix = newPrefix.replace( current_char, '-' );
         }
      }

      return getUniquePrefix( newPrefix, nsURI );
   }

   /**
    * Creates a SAAJ {@link Name} with the specified localpart, prefix, and namespace URI.
    *
    * @return a SAAJ {@link Name}
    */
   public static Name createName( String localpart,
                                  String prefix,
                                  String uri )
   throws SOAPException
   {
      LOG.debug(MSG.getMessage( Keys.CREATING_NAME,localpart,prefix,uri));
      return ( SOAPFactory.newInstance(  ).createName( localpart, prefix, uri ) );
   }

   /**
    * Converts a JAX {@link QName} to a SAAJ {@link Name}.
    *
    * @param qname a JAX {@link QName}
    *
    * @return a SAAJ {@link Name}
    */
   public static Name toName( QName qname )
   throws SOAPException
   {
      LOG.debug(MSG.getMessage( Keys.CONVRT_QNAME_TO_NAME,qname.toString()));
      return createName( qname.getLocalPart(  ),
                         qname.getPrefix(  ),
                         qname.getNamespaceURI(  ) );
   }

   /**
    * Converts a {@link javax.xml.soap.Name} to a {@link javax.xml.namespace.QName}.
    *
    * @param name a SAAJ {@link javax.xml.soap.Name}
    *
    * @return a JAX {@link javax.xml.namespace.QName}
    */
   public static QName toQName( Name name )
   {
      LOG.debug(MSG.getMessage( Keys.CONVRT_NAME_TO_QNAME,name.getQualifiedName()));
      return ( name != null ) ? new QName( name.getURI(  ),
                                           name.getLocalName(  ) ) : null;
   }

   private static String getUniquePrefix( String newprefix,
                                          String ns_uri )
   {
      //check if this prefix has been defined
      while ( s_namespace_to_prefix_map.containsValue( newprefix ) )
      {
         newprefix = newprefix + "X";
      }

      s_namespace_to_prefix_map.put( ns_uri, newprefix );

      return newprefix;
   }

   /**
    * DOCUMENT_ME
    *
    * @param current_char DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   private static boolean isValidPrefixChar( char current_char )
   {
      return ( Character.isLetterOrDigit( current_char ) ) || ( current_char == '-' ) || ( current_char == '_' )
             || ( current_char == '.' );
   }
}