/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util.ant;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.ws.util.i18n.Keys;
import org.apache.ws.util.i18n.Messages;
import org.apache.ws.util.i18n.MessagesImpl;

import java.io.File;

/**
 * LOG-DONE
 * Defines a property with a path value, allowing a base dir to be specified
 * for expanding the path if it is relative.
 *
 * @author Ian P. Springer (Hewlett-Packard Company)
 */
public class FileTask extends Task
{
   public static final Messages MSG = MessagesImpl.getInstance();
   private String m_property;
   private File m_location;
   private File m_baseDir;

   /**
    * Sets the name of the property to set.
    *
    * @param property the name of the property to set
    */
   public void setProperty( String property )
   {
      m_property = property;
   }

   /**
    * A filesystem path - may be absolute or relative. If relative it will be
    * expanded. NOTE: The parameter is a String and not a File to prevent
    * Ant from resolving relative files based on the project basedir.
    *
    * @param location a filesystem path - may be absolute or relative
    */
   public void setLocation( String location )
   {
      m_location = new File( location );
   }

   /**
    * If the specified location is relative, this directory will be used as
    * the base directory when expanding the location to an absolute path.
    * If the baseDir attribute is not specified, the project's basedir
    * (i.e. ${basedir}) will be used.
    * 
    * @param baseDir used as the base directory when expanding the location
    *                to an absolute path
    */
   public void setBaseDir( File baseDir )
   {
      m_baseDir = baseDir;
   }

   /**
    * @see {@link org.apache.tools.ant.Task#execute()}
    */
   public void execute() throws BuildException
   {
      if ( getProject() == null )
      {
         throw new IllegalStateException( MSG.getMessage( Keys.PRJ_NOT_SET) );
      }
      if ( m_property == null )
      {
         throw new BuildException( MSG.getMessage( Keys.MST_SPECIFY_PROP_NAME) );
      }
      if ( m_location == null )
      {
         throw new BuildException( MSG.getMessage( Keys.MST_SPECIFY_LOCATION) );
      }
      if ( !m_location.isAbsolute() )
      {
         m_baseDir = m_baseDir != null ? m_baseDir : getProject().getBaseDir();
         m_location = new File( m_baseDir, m_location.getPath() );
      }
      getProject().setNewProperty( m_property, m_location.getAbsolutePath() );
   }

}
