/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util.i18n;

import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.Doclet;
import com.sun.javadoc.FieldDoc;
import com.sun.javadoc.RootDoc;
import com.sun.javadoc.Tag;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Date;
import java.util.StringTokenizer;

/**
 * Generates resource.properties files from java files containing @msg Javadoc tags.
 *
 * @author Ian P. Springer (Hewlett-Packard Company)
 */
public class ResourcePropertiesGeneratorDoclet extends Doclet
{
   /**
    * The writer that is used to write the properties files.
    */
   protected static BufferedWriter s_file_out;

   /**
    * The output path for the properties files.
    */
   protected static String s_output_path;

   /**
    * A flag used to mark a file a file that contains resource keys.
    */
   protected static boolean s_is_resource_keys_file;

   /**
    * A flag used to denote when fields are found
    * in a class that have no @msg tag.  Used
    * in error checking.
    */
   protected static String s_fields_with_no_msg_tag = "";

   /**
    * The method that gets invoked to process
    * the javadocs.
    *
    * @param root  the root of the document
    *
    * @return boolean for success
    */
   public static boolean start( RootDoc root )
   {
      /*
       * Note:  One would think that if this method
       * returned false, that the javadoc system would cite an error
       * condition.  This is not true.  That is why this code
       * in this class throws exceptions instead of returning false.
       */
      boolean    result  = false;
      String     tagName = "msg";
      processOptions( root.options(  ) );
      result = processSourceFiles( root.classes(  ),
                                   tagName );
      return result;
   }

   /**
    * This method looks at each
    * of the <code>java.lang.String</code> fields for each class to see if a @msg tag exists.
    * If more than one @msg tag exists per field, it's an error.  If even one
    * <code>String</code> field contains a @msg tag, then the whole document is considered
    * a resource keys document. If any <code>String</code> fields in the document are then
    * missing @msg tags, it is considered an error.  Field names must
    * match their value, otherwise it is an error.  Non-<code>String</code> fields that do not have
    * @msg tags are ignored.  If a non-<code>String</code> field has a @msg tag, it is considered an error.
    *
    * @param classes the array of classes to be processed
    * @param tagName the name of the tag to check for
    * @throws IllegalArgumentException if there are too many @msg tags per field,
    * or if a document is considered a resource keys document, but it contains
    * fields that do not have @msg tags.
    */
   protected static boolean processSourceFiles( ClassDoc[] classes,
                                                String     tagName )
   {
      boolean result = true;

      for ( int i = 0; i < classes.length; i++ )
      {
         // get the fields of the class
         FieldDoc[] fields = classes[i].fields(  );

         for ( int j = 0; j < fields.length; j++ )
         {
            Tag[] tags = fields[j].tags( tagName );

            if ( tags.length > 1 )
            {
               result = false;
               throw new IllegalArgumentException( "Only one @" + tagName + " tag allowed per field. " + tags );
            }

            if ( tags.length > 0 )
            {
               s_is_resource_keys_file = true;
               validate( classes[i], fields[j] );
               writeToPropertiesFile( fields[j].name(  ) + "=" + tags[0].text(  ), classes[i] );
               // print a warning since ' causes problems with parameter replacement --mazz
               if ( tags[0].text(  ).indexOf( '\'' ) != -1 )
               {
                  System.err.println( "Warning: " + ResourcePropertiesGeneratorDoclet.class.getName(  )
                                      + " has detected a single quote in '" + classes[i] + "."
                                      + fields[j].name(  )
                                      + "'; make sure it does not affect parameter replacement. [" + tags[0] + "]" );
               }
            }
            else if ( tags.length == 0 )
            {
               try
               {
                  // skip it unless the data member type is a String, since that's the only type that holds a message key constant
                  Class resourceKeyClass =
                     Class.forName( classes[i].containingPackage(  ).name(  ) + "." + classes[i].name(  ) );
                  Field memberField = resourceKeyClass.getField( fields[j].name(  ) );

                  if ( memberField.getType(  ).isAssignableFrom( String.class ) )
                  {
                     s_fields_with_no_msg_tag += ( fields[j].name(  ) + " " );
                  }
               }
               catch ( Throwable t )
               {
                  // the only way for this to be thrown is if Class.forName failed or Class.getField failed.
                  // Normally this occurs because a data type is not in the classpath; in which case we should
                  // ignore this since we only care about String data types (and String is always in classpath).
                  // This may also fail with classes that have static initializers that fail; again, we typically
                  // can ignore this since resource key interfaces won't have static initializers that fail.
               }
            }
         }

         if ( s_is_resource_keys_file && ( s_fields_with_no_msg_tag.length(  ) > 0 ) )
         {
            result = false;
            closeOutputStream(  );
            throw new IllegalArgumentException( classes[i].name(  ) + " in the "
                                                + classes[i].containingPackage(  ).name(  ) + " package "
                                                + "contains a key or keys that have no associated value (or @msg tag).  "
                                                + "Please check the key(s): " + s_fields_with_no_msg_tag );
         }

         closeOutputStream(  );
         s_is_resource_keys_file     = false;
         s_fields_with_no_msg_tag    = "";
      }

      return result;
   }

   /**
    * This method makes sure that the field name and its corresponding
    * String value are identical.  If they are not identical, an
    * exception is thrown.  If the field is not of type String, an exception is thrown.
    *
    * @param classDoc is used for error reporting purposes
    * @param fieldDoc the field to be validated
    * @throws IllegalArgumentException if the field name and its String value
    * are not identical.
    * @throws RuntimeException if there is a problem with the Introspection
    * of the field value.
    * @throws SecurityException if the introspection violates a security policy
    * of the JVM.
    */
   protected static void validate( ClassDoc classDoc,
                                   FieldDoc fieldDoc )
   {
      String fieldValue = null;
      String errorMsg  = null;
      String className = classDoc.containingPackage(  ).name(  ) + "." + classDoc.name(  );

      try
      {
         Class resourceKeyClass = Class.forName( className );
         fieldValue = (String) resourceKeyClass.getField( fieldDoc.name(  ) ).get( null );
      }
      catch ( ClassCastException e )
      {
         errorMsg =
            "The class " + className + " has a field " + fieldDoc.name(  ) + " that is not of type String "
            + "but has a Javadoc tag indicating it is a resource key. "
            + "Resource key fields must be of type String.";
      }
      catch ( ClassNotFoundException e )
      {
         errorMsg =
            "The class " + className + " was not found in the classpath for the doclet. "
            + "To add this class to the classpath, make sure it is included in the -docletpath parameter. "
            + "(Doclets have their own class loaders which receive this value as their class path.)"
            + e.getMessage(  );
      }
      catch ( IllegalAccessException e )
      {
         errorMsg =
            "The class " + className + " does not allow access to the field " + fieldDoc.name(  )
            + ".  This field must be public." + e.getMessage(  );
      }
      catch ( NoSuchFieldException e )
      {
         errorMsg =
            "The compiled version of class " + className + " does not contain a field (" + fieldDoc.name(  )
            + ") that was found in the javadocs.  Make sure the compiled" + "version is up-to-date."
            + e.getMessage(  );
      }

      if ( errorMsg != null )
      {
         // this should, perhaps be a subclass of RuntimeException, however
         // nobody is going to try to make sense out of this exception, so there's
         // not much point to it.  We just want to stop processing and report a reason.
         throw new RuntimeException( errorMsg );
      }

      if ( ( fieldValue == null ) || ( !fieldValue.equals( fieldDoc.name(  ) ) ) )
      {
         throw new IllegalArgumentException( "In class " + className + ", the value of the field '"
                                             + fieldDoc.name(  ) + "' is '" + fieldValue
                                             + "', which does not equal the field name (but it must be equal)." );
      }
   }

   /**
    * This method closes the output stream to the .properties file if it is
    * open.
    */
   protected static void closeOutputStream(  )
   {
      try
      {
         if ( s_file_out != null )
         {
            s_file_out.flush(  );
            s_file_out.close(  );
            s_file_out = null;
         }
      }
      catch ( IOException e )
      {
         e.printStackTrace(  );
      }
   }

   /**
    * This method writes to the properties file.  If the file
    * is not open for writing, it opens it and prints a standard
    * header.  If the message containes new lines,
    * the white space information is stripped, before the msg is printed.
    *
    * @param msg the message to add to the properties file
    * @param sourceFile the source javadocs. needed mainly for its name.
    */
   protected static void writeToPropertiesFile( String   msg,
                                                ClassDoc sourceFile )
   {
      if ( s_file_out == null )
      {
         openOutputFile( sourceFile );
         printStandardFileHeader( sourceFile );
      }

      try
      {
         msg = stripNewLines( msg );
         s_file_out.write( msg );
         s_file_out.newLine(  );
      }
      catch ( IOException e )
      {
         e.printStackTrace(  );
      }
   }

   /**
    * Opens a properties file for writing.
    *
    * @param sourceFile the javadoc representation of a source file.
    * Used for determining the name of the properties file.
    */
   protected static void openOutputFile( ClassDoc sourceFile )
   {
      String fileName    = "resource.properties";
      String packagePath = sourceFile.containingPackage(  ).name(  ).replace( '.', File.separatorChar );
      String filePath    = s_output_path + File.separator + packagePath + File.separator + fileName;
      System.out.println( "Writing " + filePath + " ..." );
      try
      {
         s_file_out = new BufferedWriter( new FileWriter( filePath ) );
      }
      catch ( IOException e )
      {
         e.printStackTrace(  );
      }
   }

   /**
    * This method prints a standard header into the properties files.
    *
    * @param sourceFile
    */
   protected static void printStandardFileHeader( ClassDoc sourceFile )
   {
      String keysFilePath = sourceFile.qualifiedName(  ).replace( '.', File.separatorChar ) + ".java";
      try
      {
         s_file_out.write( "# Resource bundle for package " + sourceFile.containingPackage(  ).name(  ) + " and its subpackages" );
         s_file_out.newLine(  );
         s_file_out.write( "# Auto-generated " + new Date() + " from: " + keysFilePath );
         s_file_out.newLine(  );
         s_file_out.newLine(  );
         // TODO: bcp extract author name and include that (so translators know where to get answers about cryptic msgs)
      }
      catch ( IOException ioe )
      {
         ioe.printStackTrace(  );
      }
   }

   /**
    * Strips any newlines and accompanying white space from a string.
    *
    * @param message
    *
    * @return
    */
   protected static String stripNewLines( String message )
   {
      if ( message.indexOf( '\n' ) != -1 )
      {
         final StringTokenizer tokenizer = new StringTokenizer( message, "\n" );
         final StringBuffer    buf = new StringBuffer(  );

         while ( tokenizer.hasMoreTokens(  ) )
         {
            final String token = tokenizer.nextToken(  ).trim(  );

            if ( !token.equals( "" ) )
            {
               buf.append( token );
               buf.append( ' ' );
            }
         }

         message = buf.toString(  );
      }

      return ( message.trim(  ) );
   }

   /**
    * It is necessary to over-ride the super class's
    * method here because we use custom options.
    * <p>
    * -d determines the output directory.  -d was
    * chosen because it coincides with the output directory
    * for the standard doclet, which allows developers
    * to leverage certain advantages when using 3rd
    * party tools like Ant.
    * <p>
    * -p sets the flag that will store each resource bundle in its own package directory [--mazz]
    * <p>
    * If the option is unrecognized, 0 should be returned
    * to express that the option is in error.
    *
    * @param option
    *
    * @return number of params this option contains
    */
   public static int optionLength( String option )
   {
      int result = 0;
      if ( option != null )
      {
         if ( option.equals( "-d" ) )
         {
            result = 2;
         }
      }
      return result;
   }

   /**
    * From the command line options to javadoc, this method extracts
    * the output path for the properties files.
    *
    * @param options from the command line.
    */
   protected static void processOptions( String[][] options )
   {
      for ( int i = 0; i < options.length; i++ )
      {
         if ( ( options[i][0] != null ) && options[i][0].equals( "-d" ) )
         {
            s_output_path = options[i][1];
         }
      }
   }


}
