/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util.soap;

import org.apache.commons.io.CopyUtils;
import org.apache.commons.io.IOUtils;
import org.apache.ws.util.i18n.Keys;
import org.apache.ws.util.i18n.Messages;
import org.apache.ws.util.i18n.MessagesImpl;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URL;

/**
 * LOG-DONE
 * Based on SOAPClient4XG by Bob DuCharme
 * (http://www-106.ibm.com/developerworks/xml/library/x-soapcl/listing1.html). Read the SOAP envelope file passed as the
 * second parameter, pass it to the SOAP endpoint passed as the first parameter, and print out the SOAP envelope passed
 * as a response.
 *
 * Requires JDK 1.4.
 *
 * @author Bob DuCharme (IBM)
 * @author Michael Brennan (IBM)
 * @author Ian P. Springer (HP)
 */
public class SoapClient
{
   public static final Messages MSG = MessagesImpl.getInstance();
   /**
    * @param args takes three args, the third being optional: <ul> <li>serviceURL - HTTP SOAP endpoint URL to send request to
    *             <li>xmlFileToSend - a file with an XML document of the request <li>soapAction - the value for the
    *             SOAPAction HTTP header </ul>
    *
    * @throws Exception
    */
   public static void main( String[] args )
   throws Exception
   {
      validateArgs( args );
      String         serviceURL    = args[0];
      String         xmlFileToSend = args[1];
      String         soapAction    = ( args.length > 2 ) ? args[2] : "";
      String response =
         sendRequest( new URL( serviceURL ),
                      new FileInputStream( xmlFileToSend ),
                      soapAction != null ? new URI( soapAction ) : null );
      System.out.println( response );            
   }

   /**
    * @param serviceURL
    * @param soapMessageFile
    * @param soapAction
    *
    * @return
    *
    * @throws IOException
    */
   public static String sendRequest( URL         serviceURL,
                                             File        soapMessageFile,
                                             URI         soapAction )
   throws IOException
   {
      return sendRequest( serviceURL, new FileInputStream( soapMessageFile ), soapAction );
   }

   /**
    * @param serviceURL
    * @param soapMessage
    * @param soapAction
    *
    * @return
    *
    * @throws IOException
    */
   public static String sendRequest( URL         serviceURL,
                                             InputStream soapMessage,
                                             URI         soapAction )
   throws IOException
   {
      // Create the connection where we're going to send the request.
      HttpURLConnection httpConn = (HttpURLConnection) serviceURL.openConnection(  );

      // Open the input file. After we copy it to a byte array, we can see
      // how big it is so that we can set the HTTP Cotent-Length
      // property. (See complete e-mail below for more on this.)
      ByteArrayOutputStream byteOutStream = new ByteArrayOutputStream(  );

      // Copy the SOAP message to the open connection.
      CopyUtils.copy( soapMessage, byteOutStream );
      soapMessage.close(  );

      byte[] buf = byteOutStream.toByteArray(  );

      // Set the appropriate HTTP parameters.
      httpConn.setRequestProperty( "Content-Length",
                                   String.valueOf( buf.length ) );
      httpConn.setRequestProperty( "Content-Type", "text/xml; charset=UTF-8" );

      httpConn.setRequestProperty( "SOAPAction", soapAction != null ? soapAction.toString() : "" );
      httpConn.setRequestMethod( "POST" );
      httpConn.setDoOutput( true );
      httpConn.setDoInput( true );

      // Everything's set up; send the XML that was read in to buf.
      OutputStream out = httpConn.getOutputStream(  );
      out.write( buf );
      out.close(  );

      // Return the response stream as a buffered reader
      InputStream inStream;

      if ( httpConn.getResponseCode(  ) < 400 )
      {
         inStream = httpConn.getInputStream(  );
      }
      else
      {
         System.err.println( MSG.getMessage( Keys.SOAP_REQ_FAILED_HTTP_STATUS, Integer.toString(httpConn.getResponseCode(  )),
                             httpConn.getResponseMessage(  ) ));
         inStream = httpConn.getErrorStream(  );
      }

      return IOUtils.toString( inStream );
   }

   private static void validateArgs( String[] args )
   {
      String progName = SoapClient.class.getName(  );

      if ( ( args.length < 2 ) || ( args.length > 3 ) )
      {
         System.err.println(MSG.getMessage( Keys.SOAPCLIENT_USAGE,progName));
         System.err.println( MSG.getMessage( Keys.SOAPACITON_OPTIONAL) );
         System.exit( 1 );
      }
   }

}
