/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util.test.axis;

import junit.framework.Test;
import junit.framework.TestCase;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Vector;

/**
 * A JUnit test case superclass that will setup one global Axis server for use across all
 * individual test methods.  An embedded Axis server is started before the
 * first test is run and that one Axis server remains running until the last test completes
 * its run at which time the Axis server is stopped.
 *
 * @author Ian P. Springer
 */
public abstract class AbstractOneAxisTestCase
   extends AbstractAxisTestCase
{
   /** number of test methods that have been run (static since JUnit creates one instance of this class per test method) */
   private static int s_testsRun = 0;

   /** total number of test methods found in the test case (static since JUnit creates one instance of this class per test method) */
   private static int s_testCount = 0;

   /** axis server for use across all tests in the test case (static since JUnit creates one instance of this class per test method) */
   private static NotSoSimpleAxisServer s_simpleAxisServer;

   /**
    * @see AbstractAxisTestCase#AbstractAxisTestCase()
    */
   public AbstractOneAxisTestCase(  )
   {
      super(  );
   }

    /**
     * @see AbstractAxisTestCase#AbstractAxisTestCase(boolean)
     */
    protected AbstractOneAxisTestCase(boolean runServer)
    {
        super(runServer);
    }

   /**
    * @see AbstractAxisTestCase#AbstractAxisTestCase(String)
    */
   public AbstractOneAxisTestCase( String name )
   {
      super( name );
   }

   /**
    * Starts the embedded Axis server iff this is the first test being run within this test case.
    *
    * @see TestCase#setUp()
    */
   protected void setUp(  )
   throws Exception
   {
      super.setUp(  );

      //this check determines if the Axis server should be started.
      if(RUN_AXIS_SERVER)
      {
      if ( s_testsRun++ == 0 )
      {
         startAxisServer(  );
         s_simpleAxisServer = getAxisServer(  );
      }
      else
      {
         // a prior test method was run so the Axis server is already started, reuse it
         setAxisServer( s_simpleAxisServer );
      }
      }
   }

   /**
    * Stops the embedded Axis server after the last test within this test case has finished running.
    *
    * @see TestCase#tearDown()
    */
   protected void tearDown(  )
   throws Exception
   {
      super.tearDown(  );
      if ( s_testsRun == getTestCount(  ) )
      {
         //this check determines if the Axis server WAS started.
        if(RUN_AXIS_SERVER)
        {
         stopAxisServer(  );
         s_simpleAxisServer    = null;
        }
         // reset the counters to prepare for the next test case
         s_testsRun     = 0;
         s_testCount    = 0;
      }
   }

   /**
    * Returns the number of tests in this TestCase.
    *
    * @return the number of tests in this TestCase
    */
   private int getTestCount(  )
   {
      if ( s_testCount == 0 )
      {
         s_testCount = countTests( this.getClass(  ) );
      }

      return s_testCount;
   }

   /**
    * Examines the given <code>Method</code> and returns <code>true</code> if it is a JUnit test method.
    *
    * A method is considered a test method if all of the following are <code>true</code>:
    * <ol>
    *    <li>The method's name starts with "test"</li>
    *    <li>The method takes 0 parameters</li>
    *    <li>The method returns "void"</li>
    *    <li>The method is public</li>
    * </ol>
    *
    * @param m the method to check
    *
    * @return <code>true</code> if the given method is a JUnit test method; <code>false</code> otherwise
    */
   private boolean isTestMethod( Method m )
   {
      String  name       = m.getName(  );
      Class[] parameters = m.getParameterTypes(  );
      Class   returnType = m.getReturnType(  );
      boolean is_public  = Modifier.isPublic( m.getModifiers(  ) );

      return ( parameters.length == 0 ) && name.startsWith( "test" ) && returnType.equals( Void.TYPE )
             && is_public;
   }

   /**
    * Counts the test methods in the specified JUnit TestCase class.
    *
    * @param testCaseClass a JUnit TestCase class
    *
    * @return the number of test methods in the specified TestCase class
    *
    * @see #isTestMethod(Method)
    */
   private int countTests( final Class testCaseClass )
   {
      Class  superClass = testCaseClass;
      Vector tests = new Vector(  );

      while ( Test.class.isAssignableFrom( superClass ) )
      {
         Method[] methods = superClass.getDeclaredMethods(  );

         for ( int i = 0; i < methods.length; i++ )
         {
            if ( !tests.contains( methods[i].getName(  ) ) )
            {
               if ( isTestMethod( methods[i] ) )
               {
                  tests.add( methods[i].getName(  ) );
               }
            }
         }

         superClass = superClass.getSuperclass(  );
      }

      return tests.size(  );
   }
}